from click import ClickException


class CobException(Exception):
    pass


class CobExecutionError(CobException, ClickException):
    exit_code = -1


class NotInProject(CobExecutionError):
    pass


class CobConflict(CobExecutionError):
    pass


class BadMountpoint(CobConflict):
    pass


class MountpointConflict(CobConflict):
    pass



from uuid import uuid4

import functools


def task(_func=None, *, every=None, schedule=None, schedule_name=None, use_app_context=False, **kwargs):
    from .app import celery_app

    if _func is None:
        return functools.partial(task, every=every, schedule=schedule, schedule_name=schedule_name, use_app_context=use_app_context, **kwargs)


    if use_app_context:
        _func = _wrap_with_app_context(_func)

    returned = celery_app.task(**kwargs)(_func)


    if every is not None or schedule is not None:
        if every is not None and schedule is not None:
            raise RuntimeError("'every' can't be provided along with 'schedule'")

        if schedule_name is None:
            schedule_name = str(uuid4())

        celery_app.conf.beat_schedule[schedule_name] = { # pylint: disable=no-member
            'task': returned.name,
            'schedule': every if every is not None else schedule,
        }

    return returned


def _wrap_with_app_context(func):

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        from cob.app import build_app

        app = build_app(use_cached=True)
        with app.app_context():
            return func(*args, **kwargs)
    return new_func



import os

import click


@click.group()
def celery():
    pass

@celery.command(name='start-worker')
@click.argument('additional_args', nargs=-1)
def start_worker(additional_args):
    from ..bootstrapping import ensure_project_bootstrapped
    from ..project import get_project
    from ..utils.network import wait_for_app_services

    from ..app import build_app

    ensure_project_bootstrapped()
    project = get_project()
    app = build_app()
    wait_for_app_services(app)
    celery_cmd = project.build_venv_command('celery')
    tasks_subsystem = project.subsystems.tasks
    assert tasks_subsystem.grains
    argv = [celery_cmd]
    argv.extend('-A cob.celery.app worker --loglevel=DEBUG -E -B -Q {}'.format(','.join(tasks_subsystem.get_queue_names())).split())
    argv.extend(additional_args)
    os.execv(celery_cmd, argv)



import os

import click
import logbook
import yaml

from ..bootstrapping import (ensure_project_bootstrapped,
                             get_virtualenv_binary_path)
from ..project import get_project

_logger = logbook.Logger(__name__)


@click.command()
def develop():
    ensure_project_bootstrapped()
    tmuxp = get_virtualenv_binary_path('tmuxp')
    tmux_config_filename = os.path.abspath('.cob/_frontend.yml')
    with open(tmux_config_filename, 'w') as f:
        yaml.dump(_get_tmux_config(), f)
    os.execve(tmuxp, [tmuxp, 'load', tmux_config_filename], {**os.environ})


def _get_tmux_config():
    project = get_project()

    windows = [
        _window('Webapp', _project_cmd(project, 'cob testserver')),
    ]

    if project.subsystems.has_subsystem('tasks'):
        windows.extend([
            _window('Celery', _project_cmd(project, 'cob celery start-worker')),
        ])

    for subsystem in get_project().subsystems:
        subsystem.configure_tmux_window(windows)

    return {
        'session_name': 'cob-{}'.format(get_project().name),
        'windows': windows,
    }

def _window(window_name, commands, *, layout='even-horizontal'):
    if not isinstance(commands, (list, tuple)):
        commands = [commands]
    return {
        'window_name': window_name,
        'layout': layout,
        'panes': list(commands),
    }

def _project_cmd(proj, cmd):
    env = ' '.join('{}={}'.format(key, os.environ[key])
                   for key in ('COB_DEVELOP', 'COB_NO_REENTRY')
                   if key in os.environ)

    return 'cd {} && source .cob/env/bin/activate && {} {}'.format(proj.root, env, cmd)



import functools
import subprocess

from ..app import build_app
from ..exceptions import CobExecutionError
from ..bootstrapping import ensure_project_bootstrapped


def appcontext_command(func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        ensure_project_bootstrapped()
        with build_app().app_context():
            return func(*args, **kwargs)

    return new_func


def exec_or_error(*args, **kwargs):
    returned = subprocess.Popen(*args, **kwargs)
    if returned.wait() != 0:
        raise CobExecutionError('Error executing command')
    return returned



import pytest
from cob.utils.unittests import Webapp


@pytest.yield_fixture
def webapp(request):  # pylint: disable=unused-argument
    with Webapp() as _webapp, _webapp.app.app_context():
        yield _webapp



import logbook

from .base import SubsystemBase
from ..locations import Location

from flask import Blueprint

_logger = logbook.Logger(__name__)

class ViewsSubsystem(SubsystemBase):

    NAME = 'views'

    def configure_grain(self, grain, flask_app):
        _logger.trace('Found views grain: {}', grain)
        main = grain.load()
        blueprint = Blueprint(main.__name__, main.__name__, url_prefix=grain.config['mountpoint'])
        for deferred_route in getattr(main, '__cob_routes__', []):
            deferred_route.register(blueprint)
        flask_app.register_blueprint(blueprint)

    def iter_locations(self):
        assert self.grains
        for grain in self.grains:
            yield Location(grain.mountpoint)



# cob: type=tasks
from cob import task

from .models import Task, db

@task(use_app_context=True)
def task_complete(task_id):
    t = Task.query.get(task_id)
    t.completed = True
    db.session.add(t)
    db.session.commit()



import os

import logbook

from .base import SubsystemBase
from ..locations import Location

_logger = logbook.Logger(__name__)



class StaticSubsystem(SubsystemBase):

    NAME = 'static'
    SUPPORTS_OVERLAYS = True

    def configure_grain(self, grain, flask_app):
        pass

    def iter_locations(self):
        assert self.grains

        mountpoints_to_paths = {}
        for grain in self.grains:
            root = os.path.abspath(os.path.normpath(os.path.join(grain.path, grain.config['root'])))
            mountpoints_to_paths.setdefault(grain.mountpoint, []).append(root)

        for mountpoint, paths in mountpoints_to_paths.items():
            yield Location(mountpoint=mountpoint, fs_paths=paths, is_static=True)



from jinja2 import Environment, StrictUndefined
from pkg_resources import resource_string


def load_template(name):
    template_string = resource_string('cob', 'templates/{}.j2'.format(name)).decode('utf-8')
    return Environment(undefined=StrictUndefined).from_string(template_string)



"""init

Revision ID: 635350d66cc8
Revises: 
Create Date: 2017-04-18 21:25:00.509894

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '635350d66cc8'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('person',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('person')
    # ### end Alembic commands ###



import json
import os
import shutil
import socket
import subprocess
import sys
import tempfile
import time
from contextlib import contextmanager

import logbook
from jinja2 import Environment as TemplateEnvironment
from jinja2 import FileSystemLoader
import requests


_logger = logbook.Logger(__name__)

_DEFAULT_PORT = 6789

PROJECTS_ROOT = os.path.join(os.path.dirname(__file__), '_projects')


template_env = TemplateEnvironment(loader=FileSystemLoader(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), '_templates')))


_built_dockers = set()
_projs_root = tempfile.mkdtemp()


class Project(object):

    def __init__(self, name):
        super(Project, self).__init__()
        self._name = name
        self.projdir = os.path.join(_projs_root, self._name)
        if not os.path.isdir(self.projdir):
            shutil.copytree(os.path.join(
                PROJECTS_ROOT, self._name), self.projdir)

    def cmd(self, cmd, **kwargs):
        assert not cmd.startswith(
            'cob '), "You must run cob from this project's path"
        return subprocess.check_call(cmd, shell=True, cwd=self.projdir, **kwargs)

    def cob_develop_cmd(self, cmd, **kwargs):
        return self.cmd('{} {}'.format(
            os.path.join(os.path.dirname(sys.executable), 'cob'), cmd), env={'COB_DEVELOP': '1'}, **kwargs)

    def on(self, path):
        return ProjectPath(self, path)

    def _build(self):
        if self._name not in _built_dockers:
            _logger.debug('Building docker image for {._name}...', self)
            self._run_cob(['docker', 'build']).wait()
            _built_dockers.add(self._name)
        else:
            _logger.debug('Docker image for {._name} already built', self)

    @contextmanager
    def server_context(self):

        port = 8888

        self._build()
        try:
            with self._end_killing(self._run_cob(['docker', 'run', '--http-port', str(port)])) as p:
                self._wait_for_server(process=p, port=port)
                try:
                    yield RunningProject(self, port)
                except Exception:
                    self._run_cob(['docker', 'logs']).wait()
                    raise
        finally:
            self._run_cob(['docker stop']).wait()

    def _run_cob(self, argv, *, logfile=None):
        return subprocess.Popen(
            ' '.join([sys.executable, '-m', 'cob.cli.main', '-vvvvv', *argv]),
            cwd=self.projdir,
            shell=True,
            stdout=logfile,
            stderr=subprocess.STDOUT,
            env={**os.environ, **{'COB_DEVELOP': 'true'}},
        )

    @contextmanager
    def _end_killing(self, p):
        try:
            yield p
        finally:
            _logger.debug('Killing server')
            if p.poll() is None:
                p.terminate()
            p.wait()

    def _wait_for_server(self, port, timeout_seconds=30, process=None):
        end_time = time.time() + timeout_seconds
        while time.time() < end_time:
            try:
                resp = requests.get('http://127.0.0.1:{}'.format(port))
            except (socket.error, requests.ConnectionError) as e:
                if process is not None and process.poll() is not None:
                    raise RuntimeError('Process exited!')
                _logger.debug(
                    'Could not connect to server on port {} ({})', port, e)
                time.sleep(1)
                continue
            else:
                break
        else:
            _logger.error('Giving up on connection attempt to http://127.0.0.1:{}', port)
            raise RuntimeError('Could not connect')


class ProjectPath(object):

    def __init__(self, project, path):
        super(ProjectPath, self).__init__()
        self.project = project
        self.path = path

    def returns(self, code_or_string):
        resp = self._request(assert_success=False)
        if isinstance(code_or_string, int):
            assert resp.status_code == code_or_string
        else:
            assert resp.text == code_or_string
        return True

    def returns_json(self, value):
        assert self._request().json() == value
        return True

    def _request(self, **kwargs):
        with self.project.server_context() as app:
            return app.get(self.path, **kwargs)


class RunningProject(object):

    def __init__(self, project, port):
        super().__init__()
        self.project = project
        self.port = port

    def get(self, *args, **kwargs):
        return self.request('get', *args, **kwargs)

    def post(self, *args, **kwargs):
        return self.request('post', *args, **kwargs)

    def post_json(self, *args, **kwargs):
        data = json.dumps(kwargs.pop('data'))
        headers = kwargs.pop('headers', {})
        headers['Content-type'] = 'application/json'
        return self.post(*args, **kwargs, data=data, headers=headers)


    def request(self, method, path, *args, **kwargs):

        assert_success = kwargs.pop('assert_success', True)

        returned = requests.request(method, 'http://127.0.0.1:{.port}/{}'.format(self, path), *args, **kwargs)
        if assert_success:
            assert returned.ok

        return returned



# cob: mountpoint=/config
from cob import route

from flask import jsonify, current_app


@route('/')
def get_config():
    return jsonify({
        key: current_app.config[key]
        for key in [
                'SECRET_KEY',
                'SQLALCHEMY_TRACK_MODIFICATIONS',
                ]
        })



import os

import logbook

from .base import SubsystemBase
_logger = logbook.Logger(__name__)


class TasksSubsystem(SubsystemBase):

    NAME = 'tasks'


    def activate(self, flask_app):
        from ..celery.app import celery_app
        self._config = self.project.config.get('celery', {})

        # ensure critical celery config exists
        self._config.setdefault('broker_url', 'amqp://guest:guest@localhost/')

        override_broker_url = os.environ.get('COB_CELERY_BROKER_URL')
        if override_broker_url is not None:
            self._config['broker_url'] = override_broker_url

        celery_app.conf.broker_url = self._config['broker_url']
        self.queues = set()

    def get_queue_names(self):
        names = {queue_name for grain in self.grains for queue_name in grain.config.get('queue_names', [])}
        names.add('celery')
        return sorted(names)

    def configure_grain(self, grain, flask_app):
        _ = grain.load()

    def configure_app(self, flask_app):
        super().configure_app(flask_app)

        from ..celery.app import celery_app
        for task in celery_app.tasks.values():
            queue_name = getattr(task, 'queue', None)
            if queue_name is not None:
                self.queues.add(queue_name)

    def iter_locations(self):
        return None



import time

from .project import Project

def test_tasks():
    timeout_seconds = 5
    with Project('tasks').server_context() as app:
        task_id = app.get('/start').json()['id']
        end_time = time.time() + timeout_seconds
        while time.time() < end_time:
            status = app.get('/status/{}'.format(task_id)).json()['completed']
            if status:
                break
        else:
            assert False, 'Timeout expired'



import os
import click
import flask_migrate
import logbook
import multiprocessing
import random
import shutil
import string
import subprocess
from tempfile import mkdtemp

import gunicorn.app.base
from werkzeug.contrib.fixers import ProxyFix
import yaml


from ..ctx import context
from ..app import build_app
from ..bootstrapping import ensure_project_bootstrapped
from .utils import exec_or_error
from ..utils.develop import is_develop, cob_root
from ..utils.network import wait_for_app_services, wait_for_tcp
from ..utils.templates import load_template
from ..project import get_project

import pkg_resources


_COB_VERSION = pkg_resources.get_distribution('cob').version  # pylint: disable=no-member

_logger = logbook.Logger(__name__)
_CUSTOM_DOCKERFILE = "custom.docker"

def _get_user_steps():
    if not os.path.isfile(_CUSTOM_DOCKERFILE):
        return ''

    with open(_CUSTOM_DOCKERFILE) as f:
        return f.read()


@click.group()
def docker():
    pass


@docker.command()
def generate():
    proj = get_project()
    template = load_template('Dockerfile')

    if is_develop():
        sdist_file_name = _build_cob_sdist()
    else:
        sdist_file_name = None

    with open(".Dockerfile", "w") as f:
        f.write(template.render(
            project=proj,
            deployment_base_image='ubuntu:latest',
            python_version='3.6',
            is_develop=is_develop(),
            cob_sdist_filename=sdist_file_name,
            cob_root=cob_root() if is_develop() else None,
            user_steps=_get_user_steps()))


def _build_cob_sdist():
    tmpdir = mkdtemp()
    try:
        subprocess.check_call(
            'python setup.py sdist -d {}'.format(tmpdir), cwd=cob_root(), shell=True)
        [distfile] = os.listdir(tmpdir)
        returned = '.cob-sdist.tar.gz'
        shutil.move(os.path.join(tmpdir, distfile),
                    os.path.join(get_project().root, returned))
    finally:
        shutil.rmtree(tmpdir)
    return returned


@docker.command(name='build')
@click.option('--sudo', is_flag=True, help="Run docker build with sudo")
@click.option('--extra-build-args', '-e', default="", help="Arguments to pass to docker build")
def docker_build(sudo, extra_build_args):
    project = get_project()
    generate.callback()

    cmd = "{}docker build -t {} -f .Dockerfile {} .".format(
        "sudo " if sudo else "",
        project.name,
        extra_build_args)

    exec_or_error(cmd, shell=True)


@docker.command(name='wsgi-start')
def start_wsgi():
    logbook.StderrHandler(level=logbook.DEBUG).push_application()

    _ensure_secret_config()
    ensure_project_bootstrapped()
    project = get_project()
    app = build_app(config_overrides={'PROPAGATE_EXCEPTIONS': True})
    app.wsgi_app = ProxyFix(app.wsgi_app)

    wait_for_app_services(app)

    if project.subsystems.has_database():
        with app.app_context():
            if project.subsystems.models.has_migrations():
                flask_migrate.upgrade()
            else:
                context.db.create_all()

    workers_count = (multiprocessing.cpu_count() * 2) + 1

    class StandaloneApplication(gunicorn.app.base.BaseApplication):  # pylint: disable=abstract-method

        def __init__(self, app, options=None):
            self.options = options or {}
            self.application = app
            super(StandaloneApplication, self).__init__()

        def load_config(self):
            config = {key: value for key, value in self.options.items()
                      if key in self.cfg.settings and value is not None}

            for key, value in config.items():
                self.cfg.set(key.lower(), value)

        def load(self):
            return self.application

    options = {
        'bind': '0.0.0.0:8000',
        'workers': workers_count,
    }
    options.update(project.config.get('gunicorn', {}))

    StandaloneApplication(app, options).run()


def _ensure_secret_config():
    conf_dir = os.environ.get('COB_CONFIG_DIR')
    if not conf_dir:
        return
    secret_file = os.path.join(conf_dir, '000-cob-private.yml')
    if os.path.isfile(secret_file):
        return

    with open(secret_file, 'w') as f:
        f.write('flask_config:\n')
        for secret_name in ('SECRET_KEY', 'SECURITY_PASSWORD_SALT'):
            f.write('  {}: {!r}\n'.format(secret_name, _generate_secret_string()))


def _generate_secret_string(length=50):
    return "".join([random.choice(string.ascii_letters) for i in range(length)])


@docker.command(name='nginx-start')
@click.option('--print-config', is_flag=True, default=False)
def start_nginx(print_config):
    project = get_project()
    template = load_template('nginx_config')
    config = template.render({'use_ssl': False, 'hostname': None, 'project': project, 'os': os})

    if print_config:
        print(config)
        return

    wait_for_tcp('wsgi', 8000)

    with open('/etc/nginx/conf.d/webapp.conf', 'w') as f:
        f.write(config)

    nginx_path = '/usr/sbin/nginx'
    os.execv(nginx_path, [nginx_path, '-g', 'daemon off; error_log /dev/stdout info;'])


@docker.command()
@click.option('--http-port', default=None)
@click.option('--build', is_flag=True, default=False)
def run(http_port, build):
    if build:
        docker_build.callback(sudo=False, extra_build_args='')
    _exec_docker_compose(['up'], http_port=http_port)


@docker.command()
def stop():
    _exec_docker_compose(['down'])


@docker.command()
def logs():
    _exec_docker_compose(['logs'])


def _exec_docker_compose(cmd, **kwargs):
    project = get_project()
    compose_filename = '/tmp/__{}-docker-compose.yml'.format(project.name)
    with open(compose_filename, 'w') as f:
        f.write(_generate_compose_file(**kwargs))
    docker_compose = shutil.which('docker-compose')
    os.execv(docker_compose, [docker_compose, '-f',
                              compose_filename, '-p', project.name] + cmd)


@docker.command()
def compose():
    print(_generate_compose_file())


def _generate_compose_file(*, http_port=None):
    project = get_project()

    config = {
        'version': '3',
        'volumes': {
            'conf': None
        },
    }

    common_environment = {
        'COB_DATABASE_URI': 'postgresql://{0}@db/{0}'.format(project.name),
        'COB_CELERY_BROKER_URL': 'amqp://guest:guest@rabbitmq',
        'COB_CONFIG_DIR': '/conf',
    }

    services = config['services'] = {

        'wsgi':  {
            'image': project.name,
            'command': 'cob docker wsgi-start',
            'environment': common_environment,
            'depends_on': [],
            'volumes': [
                'conf:/conf',
            ],
        },

        'nginx': {
            'image': project.name,
            'command': 'cob docker nginx-start',
            'ports': ['{}:80'.format(http_port or 8000)],
            'depends_on': ['wsgi'],
        }
    }

    if project.subsystems.has_database():
        services['wsgi']['depends_on'].append('db')
        services['db'] = {
            'image': 'postgres:9.6',
            'volumes': [
                'db:/var/lib/postgresql/data'
            ],
            'environment': {
                'POSTGRES_USER': project.name,
                'POSTGRES_DB': project.name,
            }
        }
        config['volumes']['db'] = None

    if project.subsystems.has_tasks():

        services['rabbitmq'] = {
            'image': 'rabbitmq',
            #'command': 'bash -c "sleep 15 && rabbitmq-server"',
        }
        services['worker'] = {
            'image': project.name,
            'command': 'cob celery start-worker',
            'environment': common_environment,
        }

    return yaml.safe_dump(config, allow_unicode=True, default_flow_style=False)



import pytest

from cob.utils.config import merge_config, load_overrides


def test_merge_config():
    assert merge_config({'a': 'b'}, {'c': 'd'}) == {'a': 'b', 'c': 'd'}


def test_merge_config_recursive():
    a = {'a': 'b', 'c': {'x': 'y'}}
    b = {'a2': 'b2', 'c': {'x2': 'y2'}}
    result = merge_config(a, b)
    assert result == {
        'a': 'b',
        'a2': 'b2',
        'c': {'x': 'y', 'x2': 'y2'}}
    assert result is not a
    assert result['c'] is not a['c']


def test_merge_config_incompatible():
    a = {'a': 'b', 'c': 2}
    b = {'a2': 'b2', 'c': {'c': 'd2'}}
    with pytest.raises(ValueError) as caught:
        merge_config(a, b)

    assert 'cannot merge dictionaries' in str(caught.value).lower()


def test_config_overrides(tmpdir, request):
    root_config = {'a': 2}

    for i in range(3):
        with tmpdir.join('{:03}.yml'.format(i)).open('w') as f:
            f.write('a: {}'.format(1000+i))

    cfg = load_overrides(root_config, environ={'COB_CONFIG_DIR': str(tmpdir)})
    assert cfg is not root_config

    assert cfg['a'] == 1002



import functools
import os
import subprocess
import sys

import click
import logbook
import yaml

from .defs import COB_CONFIG_FILE_NAME
from .utils.develop import is_develop, cob_root
from .project import get_project

_logger = logbook.Logger(__name__)

_PREVENT_REENTRY_ENV_VAR = 'COB_NO_REENTRY'
_USE_PRE_ENV_VAR = 'COB_USE_PRE'
_PYPI_INDEX_ENV_VAR = 'COB_INDEX_URL'
_COB_REFRESH_ENV = 'COB_REFRESH_ENV'
_COB_VERSION_ENV_VAR = 'COB_VERSION'
_VIRTUALENV_PATH = '.cob/env'
_INSTALLED_DEPS = '.cob/_installed_deps.yml'

def ensure_project_bootstrapped(*, reenter=True):
    if not os.path.isfile(COB_CONFIG_FILE_NAME):
        _logger.trace('Project is not a cob project')
        return
    if _PREVENT_REENTRY_ENV_VAR in os.environ:
        _logger.trace('{} found in environ. Not reentering.', _PREVENT_REENTRY_ENV_VAR)
        return
    _ensure_virtualenv()
    if reenter:
        _reenter()

def get_virtualenv_binary_path(name):
    return os.path.join(_VIRTUALENV_PATH, 'bin', name)

def _is_in_project_virtualenv():
    venv_parent_dir = os.path.dirname(os.path.abspath(_VIRTUALENV_PATH))
    python = os.path.abspath(os.path.join(venv_parent_dir, "env", "bin", "python"))
    return os.path.abspath(sys.executable) == python

def _ensure_virtualenv():
    if not _needs_refresh():
        _logger.trace('Virtualenv already seems bootstrapped. Skipping...')
        return
    venv_parent_dir = os.path.dirname(os.path.abspath(_VIRTUALENV_PATH))
    if not _is_in_project_virtualenv():
        _logger.trace('Creating virtualenv in {}', _VIRTUALENV_PATH)
        if not os.path.isdir(venv_parent_dir):
            os.makedirs(venv_parent_dir)
        _create_virtualenv(_VIRTUALENV_PATH)

    _in_env = functools.partial(os.path.join, _VIRTUALENV_PATH, 'bin')

    if not os.path.isfile(_in_env('pip')):
        subprocess.check_call([_in_env('python'), '-m', 'ensurepip'])
    if is_develop():
        _logger.trace('Using development version of cob')
        sdist_path = os.environ.get('COB_DEVELOP_SDIST')
        if sdist_path is None:
            _virtualenv_pip_install(['-e', cob_root()])
        else:
            _virtualenv_pip_install([sdist_path])
    else:
        _logger.trace('Installing cob form Pypi')
        args = ['-U', 'cob']
        if os.environ.get(_COB_VERSION_ENV_VAR):
            version = os.environ[_COB_VERSION_ENV_VAR]
            args[-1] += '=={}'.format(version)
        if os.environ.get(_USE_PRE_ENV_VAR):
            args.append('--pre')
        if _PYPI_INDEX_ENV_VAR in os.environ:
            args.extend(['-i'], os.environ[_PYPI_INDEX_ENV_VAR])
        _virtualenv_pip_install(args)

    deps = sorted(get_project().get_deps())
    _virtualenv_pip_install(['-U', *deps])
    with open(_INSTALLED_DEPS, 'w') as f:
        yaml.dump(deps, f)

def _create_virtualenv(path):
    if 'VIRTUAL_ENV' in os.environ:
        click.echo(click.style('You are attempting to use Cob from a virtual environment. Cob will try to locate your global Python installation to avoid '
                               'unintended consequences', fg='yellow'))
        interpreter = _locate_original_interpreter()
    else:
        interpreter = sys.executable

    subprocess.check_call([interpreter, '-m', 'virtualenv', path])

def _locate_original_interpreter():
    if not os.environ.get('COB_FORCE_CURRENT_INTERPRETER'):
        for path in ('/usr/local/bin', '/usr/bin'):
            for option in ('python{0.major}.{0.minor}', 'python{0.major}'):
                optional = os.path.join(path, option.format(sys.version_info))
                if os.path.isfile(optional):
                    return optional
    else:
        click.echo(click.style('Current interpreter is forced (COB_FORCE_CURRENT_INTERPRETER is set)', fg='yellow'))

    click.echo(click.style('Could not locate global Python interpreter. Using current interpreter as fallback', fg='yellow'))
    return sys.executable

def _needs_refresh():
    if _COB_REFRESH_ENV in os.environ:
        click.echo(click.style('Virtualenv refresh forced. This might take a while...', fg='magenta'))
        return True
    if not os.path.exists(os.path.join(_VIRTUALENV_PATH, 'bin', 'python')):
        click.echo(click.style('Creating project environment. This might take a while...', fg='magenta'))
        return True
    if _get_installed_deps() != get_project().get_deps():
        click.echo(click.style('Dependencies have changes - refreshing virtualenv. This might take a while...', fg='magenta'))
        return True
    return False

def _get_installed_deps():
    if not os.path.isfile(_INSTALLED_DEPS):
        return set()
    with open(_INSTALLED_DEPS) as f:
        return set(yaml.load(f.read()))

def _virtualenv_pip_install(argv):
    _logger.trace('Installing cob in virtualenv...')
    subprocess.check_call([os.path.join(_VIRTUALENV_PATH, 'bin', 'python'), '-m', 'pip', 'install', *argv])

def _reenter():
    if _is_in_project_virtualenv():
        return

    argv = sys.argv[:]
    argv[:1] = [os.path.abspath(os.path.join(_VIRTUALENV_PATH, 'bin', 'python')), '-m', 'cob.cli.main']
    _logger.trace('Running in {}: {}...', _VIRTUALENV_PATH, argv)
    os.execve(argv[0], argv, {_PREVENT_REENTRY_ENV_VAR: 'true', **os.environ})

def _which(bin):
    for directory in os.environ['PATH'].split(':'):
        full_path = os.path.join(directory, bin)
        if os.path.isfile(full_path):
            return full_path

    raise ValueError('Could not find a python interpreter named {}'.format(bin))



import os

from .base import SubsystemBase
from ..locations import Location

class FrontendSubsystem(SubsystemBase): # pylint: disable=abstract-method
    pass

class EmberSubsystem(FrontendSubsystem):

    NAME = 'frontend-ember'

    def get_docker_pre_install_steps(self):
        return [
            'RUN npm install -g ember-cli bower',
        ]


    def get_docker_install_steps(self):
        return [
            'RUN cd {} && npm install && bower install --allow-root && ember build --environment=production'.format(grain.get_path_from('/app'))
            for grain in self.grains
        ]

    def add_grain(self, path, config):
        config.setdefault('mountpoint', '/')
        return super().add_grain(path, config)

    def configure_grain(self, grain, flask_app):
        pass

    def iter_locations(self):
        assert self.grains
        for grain in self.grains:
            yield Location(
                mountpoint=grain.mountpoint.join('assets'),
                fs_paths=[os.path.join(grain.path, 'dist/assets')],
                is_static=True,
            )
            yield Location(
                mountpoint=grain.mountpoint,
                fs_paths=[os.path.join(grain.path, 'dist/index.html')],
                is_frontend_app=self._is_location_type_auto(grain),
                is_static=True,
            )

    def _is_location_type_auto(self, grain):
        config_filename = os.path.join(grain.path, 'config', 'environment.js')
        if not os.path.isfile(config_filename):
            return False

        with open(config_filename) as f:
            for line in f:
                if 'locationType' in line:
                    return line.split(':', 1)[1].strip()[1:].lower().startswith('auto')
        return False

    def configure_tmux_window(self, windows):
        for grain in self.grains:
            windows.append({
                'window_name': 'frontend({})'.format(os.path.basename(grain.path)),
                'panes': [
                    'cd "{}" && ember build --watch'.format(grain.path),
                ],
            })



import os
import subprocess
import sys

import pytest


def test_adding_deps(tmpdir):
    assert 'COB_NO_REENTRY' not in os.environ
    with pytest.raises(ImportError):
        import pact

    projdir = tmpdir.join('proj')
    yaml = projdir.join('.cob-project.yml')
    python = str(projdir.join('.cob/env/bin/python'))


    with yaml.open('a', ensure=True) as f:
        print('name: testproj', file=f)
    _run_cob(projdir, 'bootstrap')

    assert os.path.exists(python)

    assert subprocess.call([python, '-c', 'import pact']) == 1

    with yaml.open('a') as f:
        print('deps:', file=f)
        print('  - pact', file=f)

    _run_cob(projdir, 'bootstrap')
    assert subprocess.call([python, '-c', 'import pact']) == 0


def _run_cob(cwd, cmd):

    subprocess.check_call(
        [sys.executable, '-m', 'cob.cli.main', '-vvvvv', str(cmd)],
        cwd=str(cwd),
        env={**os.environ, 'COB_DEVELOP': '1'},
    )



# pylint: disable=redefined-outer-name
import pytest

from cob.utils.url import Mountpoint

def test_mountpoint_normalize():
    assert Mountpoint('/a/b/c').path == '/a/b/c/'
    assert Mountpoint('/').path == '/'
    assert Mountpoint('').path == '/'

def test_mountpoint_join():
    assert Mountpoint('/a/b/c/').join('/').path == '/a/b/c/'
    assert Mountpoint('/a/b/c/').join('d').path == '/a/b/c/d/'
    assert Mountpoint('/a/b/c').join('/d/').path == '/a/b/c/d/'

def test_without_trailing_slash():
    assert Mountpoint('/a/b/c/').without_trailing_slash() == '/a/b/c'
    assert Mountpoint('/').without_trailing_slash() == ''

def test_str(mountpoint):
    assert str(mountpoint) == mountpoint.path

def test_hash(mountpoint):
    assert hash(mountpoint) == hash(mountpoint.path)

def test_equality(mountpoint):
    assert mountpoint == mountpoint
    assert mountpoint == Mountpoint(mountpoint.path)
    assert mountpoint == mountpoint.path
    assert not (mountpoint == Mountpoint('/bla'))
    assert not (mountpoint == '/bla')

def test_inequality(mountpoint):
    assert not (mountpoint != mountpoint)
    assert not (mountpoint != Mountpoint(mountpoint.path))
    assert not (mountpoint != mountpoint.path)
    assert mountpoint != Mountpoint('/bla')
    assert mountpoint != '/bla'



@pytest.fixture(params=['/a/b/c', '/', '/a/c/'])
def mountpoint(request):
    return Mountpoint(request.param)



from .project import Project

def test_secret_key_config():
    def _get_secret():
        with Project('basics').server_context() as app:
            return app.get('config').json()['SECRET_KEY']

    secret = _get_secret()

    assert secret
    assert len(secret) > 20

    assert _get_secret() == secret



import logbook
import gossip

from .project import get_project
# from .celery_utils import celery_app

_logger = logbook.Logger(__name__)


_cached_app = None

_building = False

def build_app(*, use_cached=False, config_overrides=None):
    from flask import Flask

    global _cached_app  # pylint: disable=global-statement
    global _building    # pylint: disable=global-statement
    if use_cached and _cached_app is not None:
        return _cached_app

    if config_overrides is None:
        config_overrides = {}

    if _building:
        raise RuntimeError('Attempted to create an app while an app was already being initialized!')

    _building = True
    try:
        proj = get_project()
        _logger.debug('Starting app {.name}...', proj)
        _cached_app = Flask(get_project().name, static_folder=None, template_folder=None)
        _cached_app.config.update(config_overrides)
        proj.configure_app(_cached_app)
        gossip.trigger_with_tags('cob.after_configure_app', {'app': _cached_app})
        _logger.trace('URL map: {}', _cached_app.url_map)
        return _cached_app
    finally:
        _building = False



import re


def ensure_trailing_slash(url):
    if not url.endswith('/'):
        url += '/'
    return url


class Mountpoint:

    def __init__(self, path):
        self.path = ensure_trailing_slash(path)

    def join(self, other):
        if not isinstance(other, Mountpoint):
            other = Mountpoint(other)
        other_path = other.path
        if other_path.startswith('/'):
            other_path = other_path[1:]
        return Mountpoint(self.path + other_path)

    def without_trailing_slash(self):
        return self.path[:-1]

    def __str__(self):
        return self.path

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other):
        if not isinstance(other, Mountpoint):
            other = Mountpoint(other)
        return self.path == other.path


def sort_paths_specific_to_generic(paths, key=None):
    return sorted(paths, key=_get_path_normalizer(key=key), reverse=True)


def _get_path_normalizer(key=None):
    if key is None:
        return _normalize_path

    def normalizer(a):
        return _normalize_path(key(a))
    return normalizer


def _normalize_path(path):

    path = re.sub(r'/+', '/', path)
    if path.startswith('/'):
        path = path[1:]
    if path.endswith('/'):
        path = path[:-1]
    return tuple(path.split('/'))



from flask_loopback import FlaskLoopback
from sentinels import NOTHING

from ..app import build_app

import json
import requests
import uuid


class Webapp(object):

    def __init__(self):
        super(Webapp, self).__init__()
        self.hostname = str(uuid.uuid1())

    def __enter__(self):
        self.activate()
        return self

    def __exit__(self, *args):
        self.deactivate()

    def activate(self):
        self.app = build_app(use_cached=True)
        self.app.config["SECRET_KEY"] = "testing_key"
        self.app.config["TESTING"] = True

        self.loopback = FlaskLoopback(self.app)
        self.loopback.activate_address((self.hostname, 80))

    def deactivate(self):
        self.loopback.deactivate_address((self.hostname, 80))

    def request(self, method, path, *args, **kwargs):
        """Performs a request against the server app
        All arguments are similar to requests' ``request``, except for a few
        exceptions listed below

        :keyword raw_response: if ``True``, returns the raw response object and avoids checking for success
        :keyword data_json: if provided, sends the value provided serialized to JSON, and sets the content type to be json accordingly
        """
        if not isinstance(path, str):
            path = str(path)
        raw_response = kwargs.pop("raw_response", False)

        data_json = kwargs.pop('data_json', NOTHING)
        if data_json is not NOTHING:
            if 'data' in kwargs:
                raise RuntimeError('`data` is not supported when combined with `data_json`')
            kwargs['data'] = json.dumps(data_json)
            headers = kwargs['headers'] = dict(kwargs.get('headers', ()))
            if not any(k.lower() == 'content-type' for k in headers):
                headers['Content-type'] = 'application/json'

        if path.startswith("/"):
            path = path[1:]
            assert not path.startswith("/")
        returned = requests.request(method, "http://{0}/{1}".format(self.hostname, path), *args, **kwargs)
        if raw_response:
            return returned

        returned.raise_for_status()
        return returned.json()

    def get(self, *args, **kwargs):
        """Shortcut for self.request('get', ...)
        """
        return self.request("get", *args, **kwargs)

    def post(self, *args, **kwargs):
        """Shortcut for self.request('post', ...)
        """
        return self.request("post", *args, **kwargs)

    def delete(self, *args, **kwargs):
        """Shortcut for self.request('delete', ...)
        """
        return self.request("delete", *args, **kwargs)

    def put(self, *args, **kwargs):
        """Shortcut for self.request('put', ...)
        """
        return self.request("put", *args, **kwargs)



import pytest

import itertools

from cob.utils.url import sort_paths_specific_to_generic


@pytest.mark.parametrize('paths', list(itertools.permutations(
    ['/api/bla', '/apibla', '/a/pi/bla', '/a'])))
def test_sort_paths(paths):
    assert sort_paths_specific_to_generic(paths) == [
        '/apibla',
        '/api/bla',
        '/a/pi/bla',
        '/a',
        ]

@pytest.mark.parametrize('paths', list(itertools.permutations(
    ['/api', '/api//blap/', '/api/blap/foo'])))
def test_sort_paths_normalization(paths):
    assert sort_paths_specific_to_generic(paths) == [
        '/api/blap/foo',
        '/api//blap/',
        '/api',
        ]



import os
import sys

import emport
import yaml
import logbook

from .defs import COB_CONFIG_FILE_NAME
from .exceptions import NotInProject
from .subsystems.manager import SubsystemsManager
from .utils.config import merge_config, load_overrides
from .utils.url import sort_paths_specific_to_generic

from flask.helpers import send_from_directory
from flask import abort

_projet = None

_logger = logbook.Logger(__name__)


DEFAULT_CONFIG = {
    'flask_config': {
        'SQLALCHEMY_TRACK_MODIFICATIONS': False,
    },
}


class Project(object):

    def __init__(self, root='.'):
        super(Project, self).__init__()
        self.root = os.path.abspath(root)
        self._static = {}

        config_filename = os.path.join(self.root, COB_CONFIG_FILE_NAME)

        if not os.path.isfile(config_filename):
            raise NotInProject('You do not seem to be in a Cob project directory')

        with open(config_filename) as f:
            config = merge_config(DEFAULT_CONFIG, yaml.load(f))

        self.config = load_overrides(config)

        self.name = self.config.get('name', os.path.basename(self.root))
        self.subsystems = SubsystemsManager(self)

        self._initialized = False

    def build_venv_command(self, cmd):
        cmd, *remainder = cmd.split()
        returned = os.path.abspath(os.path.join(self.root, '.cob', 'env', 'bin', cmd))
        return ' '.join([returned, *remainder])

    def get_deps(self):
        return set(self.config.get('deps') or ())

    def initialize(self):
        if self._initialized:
            return
        assert '_cob' not in sys.modules
        emport.set_package_name(self.root, '_cob')

        project_file_path = os.path.join(self.root, 'project.py')
        if os.path.isfile(project_file_path):
            emport.import_file(project_file_path)
        self._initialized = True

    def configure_app(self, app):
        self.initialize()
        app.config.update(self.config.get('flask_config', {}))
        self.subsystems.configure_app(app)
        self._configure_static_locations(app)

    def get_sorted_locations(self):
        return sort_paths_specific_to_generic(
            self._iter_all_locations(),
            key=lambda location: location.mountpoint.path)

    def _iter_all_locations(self):
        for subsystem in self.subsystems:
            location_iterator = subsystem.iter_locations()
            if location_iterator is None:
                continue
            yield from location_iterator

    def _configure_static_locations(self, flask_app):

        for location in self.get_sorted_locations():
            if not location.is_static():
                continue

            if location.is_frontend_app():
                flask_app.route(str(location.mountpoint), defaults={'path': location.fs_paths[0]})(
                    _static_alias_view)
            else:
                flask_app.route(str(location.mountpoint.join('<path:filename>')),
                                defaults={'search_locations': location.fs_paths})(_static_view)


def _static_view(filename, search_locations):
    for location in search_locations:
        if not location.endswith('/'):
            location += '/'
        _logger.trace('looking for {!r} in {!r}...', filename, location)
        p = os.path.join(location, filename)
        if os.path.exists(p) and os.path.abspath(p).startswith(location):
            _logger.trace('Found. Serving')
            return send_from_directory(location, filename)
    abort(404)


def _static_alias_view(path, ignored=None): # pylint: disable=unused-argument
    return send_from_directory(os.path.dirname(path), os.path.basename(path))


_project = None


def get_project():
    global _project  # pylint: disable=global-statement
    if _project is None:
        _project = Project()

    return _project



from .utils.url import Mountpoint


class Location:

    def __init__(self, mountpoint, is_frontend_app=False, is_static=False, fs_paths=()):
        assert isinstance(mountpoint, Mountpoint)
        self.mountpoint = mountpoint
        self._is_frontend_app = is_frontend_app
        self._is_static = is_static
        self.fs_paths = list(fs_paths)

    def is_frontend_app(self):
        return self._is_frontend_app

    def is_static(self):
        return self._is_static



from contextlib import closing
import time
import socket

import logbook
from urlobject import URLObject as URL


_POSTGRES_TCP_PORT = 5432
_RABBITMQ_TCP_PORT = 5672

_logger = logbook.Logger(__name__)


def wait_for_tcp(hostname, port, timeout_seconds=600):
    end_time = time.time() + timeout_seconds

    with closing(socket.socket()) as sock:
        while True:
            try:
                _logger.debug(
                    'Trying to connect to {!r}:{}...', hostname, port)
                sock.connect((hostname, port))  # pylint: disable=no-member
            except socket.error as e:
                _logger.debug('Could not connect ({})', e)
                if time.time() > end_time:
                    raise IOError(
                        'Could not connect to {}:{}'.format(hostname, port))
                time.sleep(1)
            else:
                return


def wait_for_app_services(app):
    db_uri = app.config.get('SQLALCHEMY_DATABASE_URI', None)
    if db_uri is not None:
        uri = URL(db_uri)
        if uri.scheme == 'postgresql':
            wait_for_tcp(uri.netloc.hostname, _POSTGRES_TCP_PORT)

    broker_uri = app.config.get('CELERY_BROKER_URL')
    if broker_uri is not None:
        url = URL(broker_uri)
        if url.scheme == 'amqp':
            wait_for_tcp(uri.netloc.hostname, _RABBITMQ_TCP_PORT)



from celery import Celery
from celery.loaders.base import BaseLoader


class CobLoader(BaseLoader):

    def on_worker_init(self):
        from ..app import build_app

        # this will make the tasks grains to be properly loaded and discovered
        build_app()


celery_app = Celery('cob-celery', backend='rpc://', loader=CobLoader)



import copy
import os

import yaml


def load_overrides(root, *, environ=None):
    if environ is None:
        environ = os.environ
    config_dir = environ.get('COB_CONFIG_DIR')
    if config_dir is not None and os.path.isdir(config_dir):
        for yaml_file in sorted(os.listdir(config_dir)):
            with open(os.path.join(config_dir, yaml_file)) as f:
                root = merge_config(root, yaml.load(f))
    return root

def merge_config(parent, *children):
    returned = copy.deepcopy(parent)
    for child in children:
        for key, value in child.items():
            if isinstance(value, dict):
                if key not in returned:
                    returned[key] = value.copy()
                else:
                    if not isinstance(returned[key], dict):
                        raise ValueError('Cannot merge dictionaries: value of {!r} is not a dictionary'.format(key))
                    returned[key].update(value)
            else:
                returned[key] = copy.deepcopy(value)
    return returned



# cob: type=views mountpoint=/index
from cob import route

from . import mymodels

from flask import jsonify

@route('/purge', methods=['POST'])
def purge():
    mymodels.Person.query.delete()
    mymodels.db.session.commit()
    return 'ok'

@route('/list_models')
def get_all_models():
    return jsonify([{'id': p.id} for p in mymodels.Person.query.all()])

@route('/add_model', methods=['POST'])
def add_model():
    mymodels.db.session.add(
        mymodels.Person())
    mymodels.db.session.commit()
    return 'ok'



import os
import pytest

from cob.project import Project
from cob.exceptions import MountpointConflict

from .project import PROJECTS_ROOT


def test_conflict(conflicting_project_path): # pylint: disable=redefined-outer-name
    with pytest.raises(MountpointConflict):
        Project(conflicting_project_path)


@pytest.fixture(params=os.listdir(os.path.join(PROJECTS_ROOT, 'conflicts')))
def conflicting_project_path(request):
    return os.path.join(PROJECTS_ROOT, 'conflicts', request.param)



#!/usr/bin/env python
import os
import sys

import logbook
import click


from cob.bootstrapping import ensure_project_bootstrapped

@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet", count=True)
def main(verbose, quiet):
    logbook.NullHandler().push_application()
    logbook.StreamHandler(sys.stderr, level=logbook.WARNING -
                          verbose + quiet, bubble=False).push_application()


@main.add_command
@click.command()
def bootstrap():
    ensure_project_bootstrapped()


@main.add_command
@click.command(context_settings={
    'ignore_unknown_options': True
})
@click.argument('pytest_args', nargs=-1, type=click.UNPROCESSED)
def test(pytest_args):
    ensure_project_bootstrapped()
    python_executable = os.path.abspath('.cob/env/bin/python')
    os.execv(python_executable, [python_executable, '-m', 'pytest'] + list(pytest_args))


def _add_all_subcommands():
    for name in [
            'celery',
            'db',
            'develop',
            'docker',
            'generate',
            'testserver',
            'list',
            'migrate',
    ]:
        mod = __import__('cob.cli.{}'.format(name), fromlist=[''])
        cmd = getattr(mod, name)
        main.add_command(cmd)

_add_all_subcommands()


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



import os
import logbook

from .base import SubsystemBase
from ..ctx import context

from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy


_logger = logbook.Logger(__name__)


class ModelsSubsystem(SubsystemBase):

    NAME = 'models'

    def activate(self, flask_app):

        env_override = os.environ.get('COB_DATABASE_URI')
        if env_override:
            flask_app.config['SQLALCHEMY_DATABASE_URI'] = env_override
        else:
            flask_app.config.setdefault('SQLALCHEMY_DATABASE_URI', 'sqlite:///{}'.format(os.path.join(self.project.root, '.cob', 'db.sqlite')))

        context.db = SQLAlchemy(flask_app)
        Migrate(flask_app, context.db).init_app(flask_app)
        super(ModelsSubsystem, self).activate(flask_app)

    def has_migrations(self):
        return os.path.isdir(os.path.join(self.project.root, 'migrations'))


    def configure_grain(self, grain, flask_app): # pylint: disable=unused-argument
        _logger.trace('Found models: {.path}', grain)
        grain.load()

    def iter_locations(self):
        return None



from .project import Project


def test_static_files_same_root():
    p = Project('basics')
    with p.server_context() as app:
        assert app.get('static1/file').text == 'this is #1\n'
        assert app.get('static1/file3').text == 'this is #3\n'


def test_static_files():
    p = Project('basics')
    with p.server_context() as app:
        assert app.get('static1/file').text == 'this is #1\n'
        assert app.get('static2/file').text == 'this is #2\n'



import click
import flask
import urllib

from ..app import build_app
from ..bootstrapping import ensure_project_bootstrapped


@click.group()
def list():
    pass

@list.command(name='routes')
@click.option('--server-name', default='app')
def list_routes(server_name):
    ensure_project_bootstrapped()
    app = build_app()

    app.config['SERVER_NAME'] = server_name

    with app.app_context():
        for rule in sorted(app.url_map.iter_rules(), key=lambda rule: rule.endpoint):
            options = {}
            for arg in rule.arguments:
                options[arg] = "[{0}]".format(arg)

            methods = ','.join(rule.methods)
            url = flask.url_for(rule.endpoint, **options)
            click.echo("{:50s} {:20s} ".format(rule.endpoint, methods), nl=False)
            click.echo(click.style(urllib.parse.unquote(url), fg='green'))



import re

from .project import Project


def test_ember_assets():
    with Project('ember').server_context() as app:
        for asset_uri in _iter_assets(app):
            assert app.get(asset_uri)

def _iter_assets(app):
    page = app.get('/').text
    yield from re.findall(r'<link rel="stylesheet" href="([^"]*)">', page)
    yield from re.findall(r'<script src="([^"]*)">', page)



import pytest

from .project import Project

@pytest.mark.parametrize('blueprint_type', ['directory', 'file'])
def test_blueprints(blueprint_type):
    with Project('basics').server_context() as app:
        assert app.get('blueprints/{}/test'.format(blueprint_type)).content.decode('utf-8') == 'this is {}'.format(blueprint_type)

def test_default_config_no_track_modifications():
    with Project('basics').server_context() as app:
        assert app.get('config').json()['SQLALCHEMY_TRACK_MODIFICATIONS'] == False



# cob: type=views mountpoint=/
from cob import route
from .tasks import task_complete
from .models import db, Task

from flask import jsonify

@route('start')
def start():
    task = Task()
    db.session.add(task)
    db.session.commit()
    task_complete.delay((task.id,))
    return jsonify({'id': task.id})

@route('status/<int:task_id>')
def status(task_id):
    return jsonify({'completed': Task.query.get_or_404(task_id).completed})



import subprocess
import pytest

from .project import Project

@pytest.mark.parametrize('with_migrations', [True, False])
def test_models(with_migrations):
    project_name = 'models_with_migrations' if with_migrations else 'models_no_migrations'
    with Project(project_name).server_context() as app:
        app.post('/index/purge')
        assert app.get('/index/list_models').json() == []
        app.post('/index/add_model')
        assert len(app.get('/index/list_models').json()) > 0
        assert int(subprocess.check_output(
            "docker exec {0}_db_1 psql {1} -A -t -U {1}  -c 'select count(*) from person'".format(
                project_name.replace('_', ''), project_name).strip(), shell=True)) == 1
